package dev.fitko.fitconnect.client;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.event.problems.Problem;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.AttachmentWithData;
import dev.fitko.fitconnect.api.domain.model.submission.ReceivedSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.DataIntegrityException;
import dev.fitko.fitconnect.api.exceptions.DecryptionException;
import dev.fitko.fitconnect.api.exceptions.EventCreationException;
import dev.fitko.fitconnect.api.exceptions.RestApiException;
import dev.fitko.fitconnect.api.services.Subscriber;
import dev.fitko.fitconnect.client.steps.subscriber.ConfirmSubmissionStep;
import dev.fitko.fitconnect.client.steps.subscriber.RequestSubmissionStep;
import dev.fitko.fitconnect.core.util.StopWatch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.*;

import static com.fasterxml.jackson.databind.DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildAttachmentWithData;
import static java.util.stream.Collectors.toMap;

/**
 * A fluent client for handing in a subscription for a {@link Subscriber}
 */
public final class SubscriberClient {

    private static final Logger LOGGER = LoggerFactory.getLogger(SubscriberClient.class);

    private static final ObjectMapper MAPPER = new ObjectMapper().configure(FAIL_ON_UNKNOWN_PROPERTIES, false);

    private SubscriberClient() {
    }

    public static RequestSubmissionStep builder(final Subscriber subscriber, final RSAKey privateKey) {
        if (privateKey == null) {
            LOGGER.error("No decryption key set, please check environment (config.yml) and provide a reference to the private decryption key");
            return null;
        }
        return new ClientBuilder(subscriber, privateKey);
    }

    public static class ClientBuilder implements RequestSubmissionStep, ConfirmSubmissionStep {

        private static final int DEFAULT_SUBMISSION_LIMIT = 500;

        private final Subscriber subscriber;
        private final RSAKey decryptionKey;

        private Submission submission;
        private ReceivedSubmission receivedSubmission;
        private Map<UUID, String> encryptedAttachments;
        private List<Problem> problems = new ArrayList<>();

        ClientBuilder(final Subscriber subscriber, final RSAKey privateKey) {
            this.subscriber = subscriber;
            this.decryptionKey = privateKey;
        }

        @Override
        public Set<SubmissionForPickup> getAvailableSubmissions(final UUID destinationId) {
            return getAvailableSubmissions(destinationId, 0, DEFAULT_SUBMISSION_LIMIT);
        }

        @Override
        public Set<SubmissionForPickup> getAvailableSubmissions(final UUID destinationId, final int offset, final int limit) {
            final Set<SubmissionForPickup> submissions = this.subscriber.pollAvailableSubmissions(destinationId, offset, limit);
            LOGGER.info("Received {} submission(s) for destination {}", submissions.size(), destinationId);
            return submissions;
        }

        @Override
        public List<EventLogEntry> getEventLog(final UUID caseId, final UUID destinationId) {
            return this.subscriber.getEventLog(caseId, destinationId);
        }

        @Override
        public ConfirmSubmissionStep requestSubmission(final UUID submissionId) {
            try {

                resetSavedData();

                final var startTimeDownloadSubmission = StopWatch.start();
                this.submission = this.subscriber.getSubmission(submissionId);
                LOGGER.info("Downloading submission took {}", StopWatch.stopWithFormattedTime(startTimeDownloadSubmission));

                LOGGER.info("Decrypting metadata ...");
                final Metadata metadata = decryptMetadata(this.submission.getEncryptedMetadata());
                final ValidationResult metadataValidation = this.subscriber.validateMetadata(metadata);
                if (metadataValidation.hasError()) {
                    LOGGER.error("Metadata does not match schema", metadataValidation.getError());
                    return this;
                }

                LOGGER.info("Decrypting data ...");
                final byte[] decryptedData = decryptData(this.submission.getEncryptedData());
                final String hashFromSender = getHashFromSender(metadata);
                final ValidationResult dataValidation = this.subscriber.validateHashIntegrity(hashFromSender, decryptedData);
                if (dataValidation.hasError()) {
                    LOGGER.error("Data might be corrupted, hash-sum does not match", dataValidation.getError());
                    return this;
                }

                LOGGER.info("Loading and decrypting attachments ...");
                final List<Attachment> attachmentMetadata = metadata.getContentStructure().getAttachments();
                final List<AttachmentWithData> decryptedAttachments = decryptAttachments(submissionId, attachmentMetadata);
                final ValidationResult attachmentValidation = validateAttachments(decryptedAttachments);
                if (attachmentValidation.hasError()) {
                    LOGGER.error("One of the attachments is invalid", attachmentValidation.getError());
                    return this;
                }

                this.encryptedAttachments = getEncryptedAttachments(decryptedAttachments);
                this.receivedSubmission = buildReceivedSubmission(metadata, this.submission, decryptedAttachments, decryptedData);

            } catch (final DecryptionException e) {
                LOGGER.error("Decrypting payload failed", e);
            } catch (final RestApiException e) {
                LOGGER.error("API request failed", e);
            } catch (final EventCreationException e) {
                LOGGER.error("Creating SET event failed", e);
            } catch (final IOException e) {
                LOGGER.error("Reading metadata failed", e);
            }
            return this;
        }

        @Override
        public void acceptSubmission() {
            this.subscriber.acceptSubmission(new EventPayload(this.submission, this.encryptedAttachments, this.problems));
        }

        @Override
        public void rejectSubmission(final List<Problem> rejectionProblems) {
            this.subscriber.rejectSubmission(new EventPayload(this.submission, rejectionProblems));
        }

        @Override
        public List<Problem> getProblems() {
            return this.problems;
        }

        @Override
        public Optional<ReceivedSubmission> getSubmissionData() {
            return Optional.ofNullable(this.receivedSubmission);
        }

        private List<AttachmentWithData> decryptAttachments(final UUID submissionId, final List<Attachment> attachmentMetadata) {
            if (attachmentMetadata == null || attachmentMetadata.isEmpty()) {
                LOGGER.info("Submission contains no attachments");
                return Collections.emptyList();
            }
            return decryptAttachments(getAttachments(submissionId, attachmentMetadata));
        }

        private static Map<UUID, String> getEncryptedAttachments(final List<AttachmentWithData> decryptedAttachments) {
            return decryptedAttachments.stream()
                    .collect(toMap(attachment -> attachment.getAttachmentMetadata().getAttachmentId(),
                            AttachmentWithData::getEncryptedData));
        }

        private ReceivedSubmission buildReceivedSubmission(final Metadata metadata, final Submission submission, final List<AttachmentWithData> decryptedAttachments, final byte[] decryptedData) {
            return ReceivedSubmission.builder()
                    .caseId(submission.getCaseId())
                    .destinationId(submission.getDestinationId())
                    .dataMimeType(metadata.getContentStructure().getData().getSubmissionSchema().getMimeType())
                    .attachments(decryptedAttachments)
                    .data(new String(decryptedData, StandardCharsets.UTF_8))
                    .build();
        }

        private void resetSavedData() {
            this.submission = null;
            this.receivedSubmission = null;
            this.problems = new ArrayList<>();
            this.encryptedAttachments = new HashMap<>();
        }

        private byte[] decryptData(final String encryptedData) {
            return this.subscriber.decryptStringContent(this.decryptionKey, encryptedData);
        }

        private Metadata decryptMetadata(final String encryptedMetadata) throws IOException {
            final byte[] metadataBytes = this.subscriber.decryptStringContent(this.decryptionKey, encryptedMetadata);
            return MAPPER.readValue(metadataBytes, Metadata.class);
        }

        private List<AttachmentWithData> getAttachments(final UUID submissionId, final List<Attachment> attachments) {
            final var start = StopWatch.start();
            final List<AttachmentWithData> attachmentsWithData = new ArrayList<>();
            for (final Attachment attachment : attachments) {
                final String encryptedAttachment = this.subscriber.fetchAttachment(submissionId, attachment.getAttachmentId());
                attachmentsWithData.add(buildAttachmentWithData(attachment, encryptedAttachment));
            }
            LOGGER.info("Downloading attachments took {}", StopWatch.stopWithFormattedTime(start));
            return attachmentsWithData;
        }

        private List<AttachmentWithData> decryptAttachments(final List<AttachmentWithData> encryptedAttachments) {
            final List<AttachmentWithData> decryptedAttachments = new ArrayList<>();
            for (final AttachmentWithData encryptedAttachment : encryptedAttachments) {
                final byte[] decryptedData = this.subscriber.decryptStringContent(this.decryptionKey, encryptedAttachment.getEncryptedData());
                decryptedAttachments.add(encryptedAttachment.withDecryptedData(decryptedData));
            }
            return decryptedAttachments;
        }

        private ValidationResult validateAttachments(final List<AttachmentWithData> unencryptedAttachments) {
            for (final AttachmentWithData attachmentData : unencryptedAttachments) {
                final Attachment attachmentMetadata = attachmentData.getAttachmentMetadata();
                final Hash hash = attachmentMetadata.getHash();
                final ValidationResult result = this.subscriber.validateHashIntegrity(hash.getContent(), attachmentData.getDecryptedData());
                if (result.hasError()) {
                    LOGGER.error("Attachment data for id {} is corrupted", attachmentMetadata.getAttachmentId(), result.getError());
                    return ValidationResult.error(new DataIntegrityException("Attachment " + attachmentMetadata.getAttachmentId() + " is corrupt"));
                } else {
                    LOGGER.info("Attachment {} with id {} is valid", attachmentMetadata.getFilename(), attachmentMetadata.getAttachmentId());
                }
            }
            return ValidationResult.ok();
        }

        private String getHashFromSender(final Metadata metadata) {
            return metadata.getContentStructure()
                    .getData()
                    .getHash()
                    .getContent();
        }
    }
}
